# adt-cloud-worker
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""ADT Cloud worker functional code."""


from __future__ import print_function

import argparse
import configparser
import kombu
from kombu.mixins import ConsumerMixin
import logging
import logstash
import os
import socket
import subprocess
import tarfile
import tempfile

from adt_cloud_worker._cloud_support import (
    get_default_testbed_flavor,
    set_config_dict,
    upload_tarball_to_swift_container,
)

ARCHITECTURES = (
    'amd64',
    'i386'
)
API_VERSION = "v1"


logger = logging.getLogger(__name__)


class AdtNovaWorker(ConsumerMixin):

    def __init__(self, connection, config):
        self.connection = connection
        self.config = config
        # Set the nova configuration env vars:
        _set_nova_environment_variables(self.config.items('nova'))
        # Worker is named after the machine hostname created by juju:
        # juju-<env_name>-machine-<machine_number>
        # By getting rid of the '-machine' term, the worker name is
        # identical to its corresponding security-group.
        self.worker_name = socket.gethostname().replace('-machine', '')

    def get_consumers(self, Consumer, channel):
        """Return consumers instances for all configured queues."""
        exchange = kombu.Exchange("adt.exchange", type="topic")
        queues = [
            kombu.Queue(
                'adt.requests.{}.nova.{}'.format(arch, API_VERSION),
                exchange,
                routing_key="{}.nova.{}".format(arch, API_VERSION)
            )
            for arch in ARCHITECTURES
        ]
        return [Consumer(queues=queues, callbacks=[self.process])]

    def process(self, body, message):
        """Process incomming test request.

        Run requested tests and posts results to the 'adt_results' queue
        for later checking.
        """
        body['worker'] = self.worker_name
        logger.info('Received message request: {}'.format(body), extra=body)

        # Run requested tests safely.
        tarball_path = None
        try:
            _validate_request_contents(body)
            # Create and temporary directory for storing adt results.
            dir_prefix = 'adt-{}-{}'.format(
                self.worker_name, body['request_id'])
            with tempfile.TemporaryDirectory(prefix=dir_prefix) as result_dir:
                # Run `adt-run`.
                adt_kwargs = body.copy()
                adt_kwargs['result_dir'] = result_dir
                adt_kwargs['nova_extra_args'] = self.config.get(
                    'nova', 'extra_args').split()
                adt_kwargs['nova_flavor'] = get_default_testbed_flavor()
                arguments = _make_adt_argument_list(adt_kwargs)
                logger.info(
                    'Running adt-run with: {}'.format(
                        ' '.join(arguments)), extra=body)
                body['exit_code'] = run_adt(arguments)
                logger.info(
                    'Exit code: {}'.format(body['exit_code']), extra=body)
                _create_run_metadata_file(result_dir, body)
                tarball_path = _create_tarball_from_directory(
                    result_dir,
                    'results_{}'.format(body['request_id'])
                )
                # Upload the results tarball to swift. This will be made public
                # by the adt-result-checker.
                container_name = 'adt-{}'.format(body['request_id'])
                upload_tarball_to_swift_container(tarball_path, container_name)
        except Exception as e:
            # Unexpected failures ...
            body['exit_code'] = '100'
            logger.error(e, exc_info=True, extra=body)
        finally:
            logger.info('Posting results: {}'.format(body), extra=body)
            # TODO: send error logging to result-checker in the message.
            queue = self.connection.SimpleQueue(
                'adt.results.{}'.format(API_VERSION))
            queue.put(body)
            queue.close()
            if tarball_path and os.path.exists(tarball_path):
                os.remove(tarball_path)
            message.ack()
            logger.info('Done!', extra=body)


def _create_run_metadata_file(directory, request):
    """Create a metadata file in 'directory' with some of the data from 'request'

    We want to make sure that, as (potentially confidential) items are added to
    the request that we don't blindly add them to the metadata file, so we
    have a whitelist of keys we allow.

    """
    key_whitelist = [
        'apt_pocket',
        'architecture',
        'exit_code',
        'nova_image',
        'package_name',
        'platform',
        'request_id',
        'series',
    ]
    with open(os.path.join(directory, 'nova-metadata'), 'w') as outfile:
        for key in key_whitelist:
            if key in request:
                outfile.write("{}\t{}\n".format(key, request[key]))


def _validate_request_contents(request):
    """Validates that 'request' contains everything we need to process it.

    Returns None on success, raises KeyError on failure.

    """
    required_keys = {
        'package_name',
        'nova_image',
        'request_id',
    }
    missing = required_keys.difference(set(request.keys()))
    if missing:
        raise KeyError(
            "Request is missing the following required keys: %s" % ', '.join(
                sorted(list(missing)))
            )


def run_adt(arguments):
    """Run adt-run according to the given request parameters.

    Always resets the environment with configured nova variables.

    Returns the exit code from adt-run.
    """
    cmd_line = ['adt-run'] + arguments
    proc = subprocess.Popen(
        cmd_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    proc.communicate()
    return proc.returncode


def _make_adt_argument_list(request_configuration):
    """Generate the argument list to adt-run, given the
    requestion_configuration dictionary.

    :returns: A tuple of arguments.
    """
    adt_run_args = [
        '-d',
        '--apt-source', request_configuration['package_name'],
        '--output-dir', request_configuration['result_dir'],
    ]
    if 'apt_pocket' in request_configuration:
        adt_run_args += [
            '--apt-pocket', request_configuration['apt_pocket'],
            '--apt-upgrade',
        ]
    adt_ssh_nova_args = [
        '--',
        '-d',
        '--flavor', request_configuration['nova_flavor'],
        '--image', request_configuration['nova_image'],
    ]
    if 'nova_extra_args' in request_configuration:
        adt_ssh_nova_args += request_configuration['nova_extra_args']

    return (
        adt_run_args +
        ['---', 'ssh', '-s', 'uci-nova', '-d'] +
        adt_ssh_nova_args
    )


def _set_nova_environment_variables(nova_config):
    """Set environment variables for nova configuration.

    'nova_config' will be read, and any keys that start with 'os_' will be
    set as environment variables.

    'nova_config' must be an iterable of (key, value) pairs.

    """
    # Setup nova environment variables based on the configuration file.
    for k, v in nova_config:
        if not k.startswith('os_'):
            continue
        os.environ[k.upper()] = str(v)


def _create_tarball_from_directory(directory_to_tar, tarball_root):
    """Create a gzip'd tarball of the given directory.

    :returns: the path to the newly created tarball. The caller is responsible
    for deleting the tarball.

    """
    # Compress the results directory in a temporary file.
    _unused, tarball_path = tempfile.mkstemp()
    with tarfile.open(tarball_path, "w:gz") as tar:
        for entry in os.listdir(directory_to_tar):
            tar.add(
                os.path.join(directory_to_tar, entry),
                arcname=os.path.join(tarball_root, entry),
            )
    return tarball_path


def configure_logging(config):
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)

    # Silence requests logging, which is created by nova, keystone and swift.
    requests_logger = logging.getLogger('requests')
    requests_logger.setLevel(logging.ERROR)

    # If there is no ./logs directory, fallback to stderr.
    log_path = os.path.abspath(os.path.join(__file__, '../../logs/app.log'))
    log_dir = os.path.dirname(log_path)
    if os.path.exists(log_dir):
        handler = logging.FileHandler(log_path)
    else:
        print("'logs' directory '{}' does not exist, using stderr "
              "for app log.".format(log_dir))
        handler = logging.StreamHandler()

    handler.setFormatter(
        logging.Formatter(
            '%(asctime)s  %(name)s %(levelname)s: %(message)s'
        )
    )
    root_logger.addHandler(handler)

    if 'logstash' in config:
        root_logger.addHandler(
            logstash.LogstashHandler(
                config['logstash']['host'],
                int(config['logstash']['port']),
                int(config['logstash']['version'])
            )
        )


def main():
    parser = argparse.ArgumentParser(
        description='ADT cloud worker ...')
    parser.add_argument('-c', '--conf', default='adt-service.conf',
                        help='Configuration file path')
    args = parser.parse_args()

    # Load configuration options.
    config = configparser.ConfigParser()
    config.read(args.conf)

    configure_logging(config)

    set_config_dict(config)

    amqp_uris = config.get('amqp', 'uris').split()

    with kombu.Connection(amqp_uris) as conn:
        try:
            worker = AdtNovaWorker(conn, config)
            worker.run()
        except KeyboardInterrupt:
            print('Bye!')


if __name__ == '__main__':
    main()
