# adt-cloud-worker
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Functions that talk to the cloud."""

import logging
from functools import lru_cache

from novaclient.v2.client import Client
import swiftclient.client

logger = logging.getLogger(__name__)


__all__ = [
    'get_default_testbed_flavor',
    'set_config_dict',
    'upload_tarball_to_swift_container',
]


def _nova_query_flavor():
    config = get_config_dict()
    with Client(config['nova']['os_username'], config['nova']['os_password'],
                config['nova']['os_tenant_name'],
                config['nova']['os_auth_url']) as client:
        return client.flavors.list()


@lru_cache()
def get_default_testbed_flavor(nova_query_flavor_fn=_nova_query_flavor):
    """Get the default testbed nova flavor.

    This function queries nova to determine whether the 'm1.adt.large' flavor
    has been configured, and returns an appropriate flavor name accordingly.

    """
    flavor_names = [f.name for f in nova_query_flavor_fn()]
    default = 'm1.adt.large'
    alternative = 'm1.small'
    if default in flavor_names:
        return default
    elif alternative in flavor_names:
        logging.warning(
            "Custom flavor '%s' not configured. Using "
            "'%s' instead. To reduce testbed memory consumption, "
            "configure the '%s' flavor.", default, alternative, default
        )
        return alternative
    elif flavor_names:
        first_flavor = flavor_names[0]
        logging.warning(
            "Nova has not been configured with either the default flavor (%s) "
            "of the alternative flavor (%s). Using the first reported flavor "
            "'%s' instead.", default, alternative, first_flavor
        )
        return first_flavor
    else:
        raise RuntimeError("Nova reported no flavors configured.")


def set_config_dict(config_dict):
    global get_config_dict
    get_config_dict = lambda: config_dict


get_config_dict = lambda: {}


def upload_tarball_to_swift_container(tarball_path, container_name):
    """Upload a compressed tarball to a swift container.

    The container will be created, and must not already exist.
    """
    config = get_config_dict()
    swift_client = swiftclient.client.Connection(
        authurl=config.get('nova', 'os_auth_url'),
        user=config.get('nova', 'os_username'),
        key=config.get('nova', 'os_password'),
        os_options={
            'tenant_name': config.get('nova', 'os_tenant_name'),
            'region_name': config.get('nova', 'os_region_name'),
        },
        auth_version=(
            '/v1.' in config.get('nova', 'os_auth_url') and '1.0'
            or '2.0')
    )

    swift_client.put_container(container_name)
    with open(tarball_path, 'rb') as fd:
        swift_client.put_object(
            container_name, obj='results.tgz', contents=fd,
            content_type='application/x-compressed')
