# adt-cloud-worker
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from unittest.mock import Mock

from testtools import TestCase
from testtools.matchers import raises

import fixtures
from adt_cloud_worker import _cloud_support as cs


class NovaTests(TestCase):

    def test_returns_custom_flavor_by_default(self):
        q = create_flavor_query_fn(['m1.adt.large'])
        self.assertEqual(
            'm1.adt.large',
            cs.get_default_testbed_flavor(q)
        )

    def test_returns_small_flavor_if_present(self):
        # supress warning message:
        self.useFixture(fixtures.FakeLogger())

        q = create_flavor_query_fn(['m1.smaller', 'm1.small'])
        self.assertEqual(
            'm1.small',
            cs.get_default_testbed_flavor(q))

    def test_using_nondefault_logs_error(self):
        with fixtures.FakeLogger() as f:
            q = create_flavor_query_fn(['m1.smaller', 'm1.small'])
            cs.get_default_testbed_flavor(q)
            self.assertEqual(
                "Custom flavor 'm1.adt.large' not configured. Using "
                "'m1.small' instead. To reduce testbed memory consumption, "
                "configure the 'm1.adt.large' flavor.\n",
                f.output
            )

    def test_will_use_any_flavour_if_small_not_present(self):
        # supress warning message:
        self.useFixture(fixtures.FakeLogger())

        q = create_flavor_query_fn(['m1.unknown'])
        self.assertEqual(
            'm1.unknown',
            cs.get_default_testbed_flavor(q)
        )

    def test_using_unknown_logs_error(self):
        with fixtures.FakeLogger() as f:
            q = create_flavor_query_fn(['unknown'])
            cs.get_default_testbed_flavor(q)
            self.assertEqual(
                "Nova has not been configured with either the default flavor "
                "(m1.adt.large) of the alternative flavor (m1.small). Using "
                "the first reported flavor 'unknown' instead.\n",
                f.output
            )

    def test_raises_RuntimeError_if_no_flavours_configured(self):
        q = create_flavor_query_fn([])
        self.assertThat(
            lambda: cs.get_default_testbed_flavor(q),
            raises(RuntimeError("Nova reported no flavors configured."))
        )


def create_flavor_query_fn(desired_flavour_names):
    return lambda: [
        _create_mock_flavour_object(name=n) for n in desired_flavour_names
    ]


def _create_mock_flavour_object(**kwargs):
    # Mock supports passing **kwargs in the initialiser to create arbitrary
    # attributes, but one of the arguments we need to set is 'name', which
    # is an argument to Mock itself. For that reason, we set attributes
    # individually.
    m = Mock()
    for key, value in kwargs.items():
        setattr(m, key, value)
    return m
