# adt-cloud-worker
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import tempfile
import tarfile

from fixtures import (
    Fixture,
    TempDir,
)
import testtools
from testtools.matchers import (
    FileExists,
    TarballContains,
    raises,
)

from adt_cloud_worker import (
    _create_tarball_from_directory,
    _make_adt_argument_list,
    _set_nova_environment_variables,
    _validate_request_contents,
    _create_run_metadata_file,
)


class TestADTArgumentGeneration(testtools.TestCase):

    def createRequiredAdtRequestArgumets(self, **kwargs):
        """Generate a dict that contains all the required parts of the
        adt-cloud-worker request.

        You may pass in additional arguments to customise the contents of the
        dictionary.
        """
        defaults = {
            'worker': 'worker-foo',
            'package_name': "some_package_name",
            'result_dir': '/tmp/some/dir',
            'nova_flavor': 'some-nova-flavour',
            'nova_image': 'some-nova-image',
        }
        defaults.update(kwargs)
        return defaults

    def test_required_arguments(self):
        observed = _make_adt_argument_list(
            self.createRequiredAdtRequestArgumets()
        )
        self.assertEqual(
            [
                '-d',
                '--apt-source', 'some_package_name',
                '--output-dir', '/tmp/some/dir',
                '---',
                'ssh',
                '-s', 'uci-nova',
                '-d',
                '--',
                '-d',
                '--flavor', 'some-nova-flavour',
                '--image', 'some-nova-image',
            ],
            observed)

    def test_includes_pocket_args(self):
        observed = _make_adt_argument_list(
            self.createRequiredAdtRequestArgumets(apt_pocket='foo')
        )
        self.assertEqual(
            [
                '-d',
                '--apt-source', 'some_package_name',
                '--output-dir', '/tmp/some/dir',
                '--apt-pocket', 'foo', '--apt-upgrade',
                '---',
                'ssh',
                '-s', 'uci-nova',
                '-d',
                '--',
                '-d',
                '--flavor', 'some-nova-flavour',
                '--image', 'some-nova-image',
            ],
            observed)


class RestoreEnvironmentVariableFixture(Fixture):

    def __init__(self, env_var_name):
        self.name = env_var_name

    def setUp(self):
        # TODO: Make this python3-ised
        super(RestoreEnvironmentVariableFixture, self).setUp()
        old_value = os.environ.get(self.name)
        self.addCleanup(self._restore_value,  old_value)

    def _restore_value(self, value):
        if value is None:
            try:
                del os.environ[self.name]
            except KeyError:
                # key was never set, nbothing to do:
                pass
        else:
            os.environ[self.name] = value


class NovaEnvironmentVariableTests(testtools.TestCase):

    def test_does_nothing_with_empty_dict(self):
        _set_nova_environment_variables({})

    def test_sets__os_keys(self):
        with RestoreEnvironmentVariableFixture('OS_FOO'):
            _set_nova_environment_variables([('os_foo', 'bar')])
            self.assertEqual('bar', os.getenv('OS_FOO'))

    def test_ignores_non_os_keys(self):
        with RestoreEnvironmentVariableFixture('FOO'):
            os.environ['FOO'] = 'bar'
            _set_nova_environment_variables([('foo', 'baz')])
            self.assertEqual('bar', os.getenv('FOO'))

    def test_stringifies_values(self):
        with RestoreEnvironmentVariableFixture('OS_FOO'):
            _set_nova_environment_variables([('os_foo', 123)])
            self.assertEqual('123', os.getenv('OS_FOO'))


class TarballCreationTests(testtools.TestCase):

    def create_tarball_with_cleanup(self, directory_to_tar, tarball_root):
        tarball_path = _create_tarball_from_directory(directory_to_tar, tarball_root)
        self.addCleanup(os.remove, tarball_path)
        return tarball_path

    def test_will_tar_empty_directory(self):
        with tempfile.TemporaryDirectory() as tempdir:
            tarfile_path = self.create_tarball_with_cleanup(tempdir, 'root')

            self.assertThat(tarfile_path, FileExists())
            self.assertTrue(tarfile.is_tarfile(tarfile_path))

    def test_tarball_includes_directory_name(self):
        with tempfile.TemporaryDirectory() as tempdir:
            with open(os.path.join(tempdir, 'testfile'), 'w') as f:
                f.write("Some Data")
            tarfile_path = self.create_tarball_with_cleanup(tempdir, 'root')

            expected = 'root/testfile'
            self.assertThat(
                tarfile_path,
                TarballContains([expected])
            )


class TestRequestValidation(testtools.TestCase):

    def assertRequestFailsWithError(self, request, expected_error):
        self.assertThat(
            lambda: _validate_request_contents(request),
            raises(KeyError(expected_error))
        )

    def test_empty_request(self):
        self.assertRequestFailsWithError(
            {},
            'Request is missing the following required keys: nova_image, '
            'package_name, request_id'
        )

    def test_one_missing(self):
        self.assertRequestFailsWithError(
            dict(nova_image='foo', request_id='123'),
            'Request is missing the following required keys: package_name'
        )

    def test_all_present(self):
        _validate_request_contents(
            dict(nova_image='foo', request_id='123', package_name='librepng')
        )


class TestMetadataFileCreation(testtools.TestCase):

    def assertRequestResultsInFileContents(self, request, expected_contents):
        with TempDir() as t:
            _create_run_metadata_file(t.path, request)
            expected_path = os.path.join(t.path, 'nova-metadata')

            self.assertThat(
                expected_path,
                FileExists()
            )
            with open(expected_path) as infile:
                contents = infile.read()
            self.assertEquals(contents, expected_contents)

    def test_ignores_non_whitelisted_keys(self):
        self.assertRequestResultsInFileContents(
            dict(secret="Wouldn't you like to know!"),
            ''
        )

    def test_writes_exit_code(self):
        self.assertRequestResultsInFileContents(
            dict(exit_code='100'),
            'exit_code\t100\n'
        )

    def test_writes_multiple_whitelisted_keys(self):
        self.assertRequestResultsInFileContents(
            dict(series='trusty', exit_code='100'),
            'exit_code\t100\nseries\ttrusty\n'
        )
